# -*- coding: utf-8 -*-

# Copyright 2017 IBM RESEARCH. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

"""Standard gates."""
from .barrier import barrier
from .ccx import ccx
from .cswap import cswap
from .cx import cx
from .cxbase import cx_base
from .cy import cy
from .cz import cz
from .swap import swap
from .h import h
from .iden import iden
from .s import s
from .t import t
from .u0 import u0
from .u1 import u1
from .u2 import u2
from .u3 import u3
from .ubase import u_base
from .x import x
from .y import y
from .z import z
from .rx import rx
from .ry import ry
from .rz import rz
from .cu1 import cu1
from .ch import ch
from .crz import crz
from .cu3 import cu3
